-- This software is distributed under the terms of the new BSD License.
-- Copyright (c) 2014 molelord
-- All rights reserved.

-- 書籍「CPUの創り方」のCPUをVHDLで実装してみたもの。
-- Clock EnableポートはオリジナルのTD4には存在しないが、それがないと
-- 10Hzや1Hzでの動作のために分周器で作ったクロックが必要になってしまう
-- ため、付け足した。

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.td4inst_pkg.all;

entity td4core is
port (
    clock_i   : in  std_logic;
    reset_n_i : in  std_logic;
    ce_i      : in  std_logic; -- Clock Enable
    d_i       : in  std_logic_vector(7 downto 0); -- ROMのデータバス
    a_o       : out std_logic_vector(3 downto 0); -- ROMのアドレスバス
    iport_i   : in  std_logic_vector(3 downto 0);
    oport_o   : out std_logic_vector(3 downto 0)
);
end td4core;

architecture rtl of td4core is

    signal a_reg    : std_logic_vector(3 downto 0);
    signal b_reg    : std_logic_vector(3 downto 0);
    signal oport    : std_logic_vector(3 downto 0);
    signal pcounter : std_logic_vector(3 downto 0);
    signal c_flag   : std_logic;

    signal opcode   : std_logic_vector(3 downto 0);
    signal imm      : std_logic_vector(3 downto 0);

    signal load_a_reg    : std_logic;
    signal load_b_reg    : std_logic;
    signal load_oport    : std_logic;
    signal load_pcounter : std_logic;
    signal sel           : std_logic_vector(1 downto 0);
    signal selout        : std_logic_vector(3 downto 0);
    signal aluresult     : unsigned(4 downto 0);
    signal aluout        : unsigned(3 downto 0);
    signal carry         : std_logic;

    constant SEL_A     : std_logic_vector(1 downto 0) := "00";
    constant SEL_B     : std_logic_vector(1 downto 0) := "01";
    constant SEL_IPORT : std_logic_vector(1 downto 0) := "10";
    constant SEL_ZERO  : std_logic_vector(1 downto 0) := "11";

begin

    oport_o <= oport;
    a_o     <= pcounter;

    process (clock_i, reset_n_i) begin
        if reset_n_i = '0' then
            a_reg <= (others=>'0');
        elsif clock_i'event and clock_i = '1' then
            if ce_i = '1' then
                if load_a_reg = '1' then
                    a_reg <= std_logic_vector(aluout);
                end if;
            end if;
        end if;
    end process;

    process (clock_i, reset_n_i) begin
        if reset_n_i = '0' then
            b_reg <= (others=>'0');
        elsif clock_i'event and clock_i = '1' then
            if ce_i = '1' then
                if load_b_reg = '1' then
                    b_reg <= std_logic_vector(aluout);
                end if;
            end if;
        end if;
    end process;

    process (clock_i, reset_n_i) begin
        if reset_n_i = '0' then
            oport <= (others=>'0');
        elsif clock_i'event and clock_i = '1' then
            if ce_i = '1' then
                if load_oport = '1' then
                    oport <= std_logic_vector(aluout);
                end if;
            end if;
        end if;
    end process;

    process (clock_i, reset_n_i) begin
        if reset_n_i = '0' then
            pcounter <= (others=>'0');
        elsif clock_i'event and clock_i = '1' then
            if ce_i = '1' then
                if load_pcounter = '1' then
                    pcounter <= std_logic_vector(aluout);
                else
                    pcounter <= std_logic_vector(unsigned(pcounter) + 1);
                end if;
            end if;
        end if;
    end process;

    process (clock_i, reset_n_i) begin
        if reset_n_i = '0' then
            c_flag <= '0';
        elsif clock_i'event and clock_i = '1' then
            if ce_i = '1' then
                c_flag <= carry;
            end if;
        end if;
    end process;

    opcode <= d_i(7 downto 4);
    imm    <= d_i(3 downto 0);

    -- セレクタ
    process (sel, a_reg, b_reg, iport_i) begin
        case sel is
            when SEL_A =>
                selout <= a_reg;
            when SEL_B =>
                selout <= b_reg;
            when SEL_IPORT =>
                selout <= iport_i;
            when SEL_ZERO =>
                selout <= (others=>'0');
            when others =>
                selout <= (others=>'0');
        end case;
    end process;

    -- ALU
    aluresult <= unsigned("0" & selout) + unsigned("0" & imm);
    carry     <= aluresult(aluresult'left);
    aluout    <= aluresult(aluresult'left - 1 downto 0);

    -- デコーダ
    process (opcode, c_flag) begin
        load_a_reg    <= '0';
        load_b_reg    <= '0';
        load_oport    <= '0';
        load_pcounter <= '0';
        sel           <= SEL_A;
        case opcode is
            when MOV_A_IM =>
                load_a_reg <= '1';
                sel        <= SEL_ZERO;
            when MOV_B_IM =>
                load_b_reg <= '1';
                sel        <= SEL_ZERO;
            when MOV_A_B  =>
                load_a_reg <= '1';
                sel        <= SEL_B;
            when MOV_B_A  =>
                load_b_reg <= '1';
                sel        <= SEL_A;
            when ADD_A_IM =>
                load_a_reg <= '1';
                sel        <= SEL_A;
            when ADD_B_IM =>
                load_b_reg <= '1';
                sel        <= SEL_B;
            when IN_A =>
                load_a_reg <= '1';
                sel        <= SEL_IPORT;
            when IN_B =>
                load_b_reg <= '1';
                sel        <= SEL_IPORT;
            when OUT_IM =>
                load_oport <= '1';
                sel        <= SEL_ZERO;
            when OUT_B =>
                load_oport <= '1';
                sel        <= SEL_B;
            when JMP_IM =>
                load_pcounter <= '1';
                sel           <= SEL_ZERO;
            when JNC_IM =>
                sel        <= SEL_ZERO;
                if c_flag = '0' then
                    load_pcounter <= '1';
                end if;
            when others =>
                null;
        end case;
    end process;

end RTL;
