-- This software is distributed under the terms of the new BSD License.
-- Copyright (c) 2014 molelord
-- All rights reserved.

-- td4coreをSpartan-6 LX9 MicroBoardで動かすための周辺回路。

-- DIPSWの4番はクロック周波数の切り替えに使用して、3,2,1番を入力ポート
-- につなぐので、TD4の入力ポートのbit3は常にゼロである。

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.compo_pkg.all;
use work.td4inst_pkg.all;

entity td4shell is
port (
    user_clock_i : in  std_logic; -- 40Mhz
    user_reset_i : in  std_logic; -- スイッチ直結なのでバウンスあり
    gpio_dip_i   : in  std_logic_vector(4 downto 1);
    gpio_led_o   : out std_logic_vector(4 downto 1)
);
end td4shell;

architecture rtl of td4shell is

    signal clk   : std_logic;
    signal rst   : std_logic;
    signal rst_n : std_logic;
    signal iport : std_logic_vector(3 downto 0);
    signal data  : std_logic_vector(7 downto 0);
    signal addr  : std_logic_vector(3 downto 0);

    signal gpio_dip    : std_logic_vector(4 downto 1);
    signal gpio_dip_ff : std_logic_vector(4 downto 1);

    subtype romline is std_logic_vector(7 downto 0);
    type rom is array (0 to 15) of romline;

    -- サンプルプログラム1 : LEDちかちか
    -- constant INSTROM : rom := (
    --    OUT_IM   & "0011", -- 0
    --    OUT_IM   & "0110", -- 1
    --    OUT_IM   & "1100", -- 2
    --    OUT_IM   & "1000", -- 3
    --    OUT_IM   & "1000", -- 4
    --    OUT_IM   & "1100", -- 5
    --    OUT_IM   & "0110", -- 6
    --    OUT_IM   & "0011", -- 7
    --    OUT_IM   & "0001", -- 8
    --    JMP_IM   & "0000", -- 9
    --    ADD_A_IM & "0000", -- a
    --    ADD_A_IM & "0000", -- b
    --    ADD_A_IM & "0000", -- c
    --    ADD_A_IM & "0000", -- d
    --    ADD_A_IM & "0000", -- e
    --    ADD_A_IM & "0000"  -- f
    --);

    -- サンプルプログラム2 : ラーメンタイマー
    constant INSTROM : rom := (
        OUT_IM   & "0111", -- 0
        ADD_A_IM & "0001", -- 1
        JNC_IM   & "0001", -- 2
        ADD_A_IM & "0001", -- 3
        JNC_IM   & "0011", -- 4
        OUT_IM   & "0110", -- 5
        ADD_A_IM & "0001", -- 6
        JNC_IM   & "0110", -- 7
        ADD_A_IM & "0001", -- 8
        JNC_IM   & "1000", -- 9
        OUT_IM   & "0000", -- a
        OUT_IM   & "0100", -- b
        ADD_A_IM & "0001", -- c
        JNC_IM   & "1010", -- d
        OUT_IM   & "1000", -- e
        JMP_IM   & "1111"  -- f
    );

    constant CLK_HZ : integer := 40000000;
    signal count_max : integer range 0 to CLK_HZ - 1;
    signal count     : integer range 0 to CLK_HZ - 1;
    signal tick      : std_logic;

begin

    iport <= "0" & gpio_dip_ff(3 downto 1);

    -- 「CPUの創りかた」の回路と異なり、基板上にCR+シュミットトリガが
    -- ついているわけではないため、デバウンスしたものをリセット信号とする。
    deb_rst : debounce_rst
    generic map (
        INIT   => '0',
        CLK_HZ => CLK_HZ
    )
    port map (
        clk_i => user_clock_i,
        rst_i => user_reset_i,
        rst_o => rst
    );
    rst_n <= not rst;


    -- clkを10Hzまたは1Hzでトグルさせるのが目的。
    -- DIPSW(4)がHighなら10Hz、Lowなら1Hzにする。
    count_max <= CLK_HZ/10 - 1 when gpio_dip_ff(4) = '1' else
                 CLK_HZ/1 - 1;

    process (user_clock_i, rst_n) begin
        if rst_n = '0' then
            count <= 0;
        elsif user_clock_i'event and user_clock_i = '1' then
            if count < count_max then
                count <= count + 1;
            else
                count <= 0;
            end if;
        end if;
    end process;

    tick <= '1' when count = count_max else '0';

    clk <= user_clock_i;

    -- 非同期入力なのでDobule flopsする。
    process (user_clock_i, rst_n) begin
        if rst_n = '0' then
            gpio_dip    <= (others=>'0');
            gpio_dip_ff <= (others=>'0');
        elsif user_clock_i'event and user_clock_i = '1' then
            gpio_dip    <= gpio_dip_i;
            gpio_dip_ff <= gpio_dip;
        end if;
    end process;


    cpu : td4core
    port map (
        clock_i   => clk,
        reset_n_i => rst_n,
        ce_i      => tick,
        d_i       => data,
        a_o       => addr,
        iport_i   => iport,
        oport_o   => gpio_led_o
    );

    data <= INSTROM(to_integer(unsigned(addr)));

end RTL;
